unit mozreg;

interface
uses
  Classes;

type
  TMozillaNode=class
  private
    Location, NameLocation, Left, AParent: Longint;
    NameLength: SmallInt;
    //Key node
    SubKey, Entry, None: Longint;
    //Entry node
    ValueLength, ValueLocation: Longint;
    //Hleper for saving
    SubKeyLocation, EntryLocation, LeftLocation: Longint;
    procedure SaveNode(F: TStream);
  public
    AType: SmallInt;
    Name, ValueString: AnsiString;
    ValueInt: Longint;
    ValueByte: Array of byte;
    ValueFile: AnsiString; //I do not know the type
    Parent: TMozillaNode;
    Entries, Keys: TList;
    constructor Create(AParent: TMozillaNode);
    destructor Destroy; override;
    procedure Load(F: TStream; Pos: Longint);
    procedure SaveChilds(F: TStream);
    function FindName(const AName: string): TMozillaNode;
  end;

  TMozillaRegistry=class
  private
    F: TStream;
    procedure LoadHeader;
    procedure SaveHeader;
  public
    Header: longint;
    MajorVersion, MinorVersion: SmallInt;
    NextAvail, RootNode, NextAvailPosition: Longint;
    Root: TMozillaNode;
    procedure Load(const FileName:string);
    procedure Save(const FileName:string);
  end;

const
  MOZILLA_REG_USERS_ROOT_KEY=1;
  MOZILLA_REG_COMMON_ROOT_KEY=2;
  MOZILLA_REG_CURRENT_USER_ROOT_KEY=3;
  MOZILLA_REG_PRIVATE_ROOT_KEY=4;

  MOZILLA_REG_ENTRIES=$10;
  MOZILLA_REG_UTF_STRING_ENTRY=MOZILLA_REG_ENTRIES+1;
  MOZILLA_REG_32_INTEGER_ENTRY=MOZILLA_REG_ENTRIES+2;
  MOZILLA_REG_BYTE_ENTRY=MOZILLA_REG_ENTRIES+3;
  MOZILLA_REG_FILE_ENTRY=MOZILLA_REG_ENTRIES+4;

implementation

uses
  SysUtils;

{ TMozillaNode }

constructor TMozillaNode.Create(AParent: TMozillaNode);
begin
  Parent:=AParent;
  Entries:=TList.Create;
  Keys:=TList.Create;
end;

destructor TMozillaNode.Destroy;
begin
  Entries.Free;
  Keys.Free;
  inherited;
end;

function TMozillaNode.FindName(const AName: string): TMozillaNode;
var
  i: integer;
begin
  Result:=nil;
  for i:=0 to Entries.Count-1 do begin
    if TMozillaNode(Entries.Items[i]).Name=AName then begin
      Result:=TMozillaNode(Entries.Items[i]);
      Exit;
    end;
  end;
  for i:=0 to Keys.Count-1 do begin
    Result:=TMozillaNode(Keys.Items[i]).FindName(AName);
    if Assigned(Result) then
      Break;
  end;
end;

procedure TMozillaNode.Load(F: TStream; Pos: Longint);
var
  P: Pointer;
  Node: TMozillaNode;
begin
  F.Seek(Pos, soFromBeginning);
  F.Read(Location, SizeOf(Location));

  F.Read(NameLocation, SizeOf(NameLocation));
  F.Read(NameLength, SizeOf(NameLength)); //With null terminator
  F.Read(AType, SizeOf(AType));
  F.Read(Left, SizeOf(Left)); //0 - no next sibling
  if AType>MOZILLA_REG_ENTRIES then begin //Entries
    F.Read(ValueLength, SizeOf(ValueLength));
    F.Read(ValueLocation, SizeOf(ValueLocation));
    F.Read(ValueLength, SizeOf(ValueLength));
  end else begin //Keys
    F.Read(SubKey, SizeOf(SubKey)); //First subkey
    F.Read(Entry, SizeOf(Entry)); //First entry
    F.Read(None, SizeOf(None));
  end;
  F.Read(AParent, SizeOf(AParent)); //0 - no parent

  F.Seek(NameLocation, soFromBeginning);
  SetLength(Name, NameLength-1);
  P:=PChar(Name);
  F.Read(P^, NameLength);
  if AType>MOZILLA_REG_ENTRIES then begin
    F.Seek(ValueLocation, soFromBeginning);
    case AType of
      MOZILLA_REG_UTF_STRING_ENTRY, MOZILLA_REG_FILE_ENTRY: begin
        SetLength(ValueString, ValueLength-1);
        P:=PChar(ValueString);
        F.Read(P^, ValueLength);
      end;
      MOZILLA_REG_32_INTEGER_ENTRY:
        F.Read(ValueInt, ValueLength);
      MOZILLA_REG_BYTE_ENTRY: begin
        SetLength(ValueByte, ValueLength);
        F.Read(ValueByte[0], ValueLength);
      end;
    end;
  end;
  if Left<>0 then begin
    Node:=TMozillaNode.Create(Parent);
    Node.Load(F, Left);
    if Node.AType>MOZILLA_REG_ENTRIES then
      Parent.Entries.Add(Node)
    else
      Parent.Keys.Add(Node);
  end;
  if Entry<>0 then begin
    Node:=TMozillaNode.Create(Self);
    Node.Load(F, Entry);
    if Node.AType>MOZILLA_REG_ENTRIES then
      Entries.Add(Node)
    else
      Keys.Add(Node);
  end;
  if SubKey<>0 then begin
    Node:=TMozillaNode.Create(Self);
    Node.Load(F, SubKey);
    if Node.AType>MOZILLA_REG_ENTRIES then
      Entries.Add(Node)
    else
      Keys.Add(Node);
  end;
end;

procedure TMozillaNode.SaveChilds(F: TStream);
var
  i:integer;
  PrevNode, Node: TMozillaNode;
begin
  //first save entries
  PrevNode:=nil;
  for i:=0 to Entries.Count-1 do begin
    Node:=TMozillaNode(Entries.Items[i]);
    Node.SaveNode(F);
    if Assigned(PrevNode) then begin
      Node.Left:=Node.Location;
      F.Seek(PrevNode.LeftLocation, soFromBeginning);
      F.Write(Node.Left, SizeOf(Node.Left));
    end else begin
      Node.Entry:=Node.Location;
      F.Seek(EntryLocation, soFromBeginning);
      F.Write(Node.Entry, SizeOf(Node.Entry));
    end;
    PrevNode:=Node;
    F.Seek(0, soFromEnd);
  end;

  //next save keys
  PrevNode:=nil;
  for i:=0 to Keys.Count-1 do begin
    Node:=TMozillaNode(Keys.Items[i]);
    Node.SaveNode(F);
    if Assigned(PrevNode) then begin
      Node.Left:=Node.Location;
      F.Seek(PrevNode.LeftLocation, soFromBeginning);
      F.Write(Node.Left, SizeOf(Node.Left));
    end else begin
      Node.SubKey:=Node.Location;
      F.Seek(SubkeyLocation, soFromBeginning);
      F.Write(Node.SubKey, SizeOf(Node.SubKey));
    end;
    PrevNode:=Node;
    F.Seek(0, soFromEnd);
  end;

  for i:=0 to Keys.Count-1 do begin
    Node:=TMozillaNode(Keys.Items[i]);
    Node.SaveChilds(F);
  end;

end;

procedure TMozillaNode.SaveNode(F: TStream);
var
  s: Array of char;
begin
  SubKey:=0; //Zero links
  Entry:=0;
  Left:=0;

  NameLocation:=F.Position;
  NameLength:=Length(Name)+1;
  SetLength(s, NameLength);
  StrPCopy(@s[0], Name);
  F.Write(s[0], NameLength);
  if AType>MOZILLA_REG_ENTRIES then begin
    ValueLocation:=F.Position;
    case AType of
      MOZILLA_REG_UTF_STRING_ENTRY, MOZILLA_REG_FILE_ENTRY: begin
        ValueLength:=Length(ValueString)+1;
        SetLength(s, ValueLength);
        StrPCopy(@s[0], ValueString);
        F.Write(s[0], ValueLength);
      end;
      MOZILLA_REG_32_INTEGER_ENTRY: begin
        ValueLength:=SizeOf(ValueInt);
        F.Write(ValueInt, ValueLength);
      end;
      MOZILLA_REG_BYTE_ENTRY: begin
        ValueLength:=High(ValueByte)+1;
        F.Write(ValueByte[0], ValueLength);
      end;
    end;
  end;
  Location:=F.Position;
  F.Write(Location, SizeOf(Location));
  F.Write(NameLocation, SizeOf(NameLocation));
  F.Write(NameLength, SizeOf(NameLength));
  F.Write(AType, SizeOf(AType));
  LeftLocation:=F.Position;
  F.Write(Left, SizeOf(Left)); //Prozatim nevim, az ho budu vedet tak prepisu
  if AType>MOZILLA_REG_ENTRIES then begin
    F.Write(ValueLength, SizeOf(ValueLength));
    F.Write(ValueLocation, SizeOf(ValueLocation));
    F.Write(ValueLength, SizeOf(ValueLength));
  end else begin //Keys
    SubKeyLocation:=F.Position;
    F.Write(Subkey, SizeOf(Subkey)); //Prozatim nevim, az ho budu vedet tak prepisu
    EntryLocation:=F.Position;
    F.Write(Entry, SizeOf(Entry)); //Prozatim nevim, az ho budu vedet tak prepisu
    None:=0;
    F.Write(None, SizeOf(None));
  end;
  if Assigned(Parent) then
    AParent:=Parent.Location
  else
    AParent:=0;
  F.Write(AParent, SizeOf(AParent));
end;

{ TMozillaRegistry }

procedure TMozillaRegistry.Load(const FileName: string);
begin
  F:=TFileStream.Create(FileName, fmOpenRead);
  try
    LoadHeader;
    Root:=TMozillaNode.Create(nil);
    Root.Load(F, RootNode);
  finally
    F.Free;
  end;
end;

procedure TMozillaRegistry.LoadHeader;
begin

  F.read(Header, SizeOf(Header));
  if Header<>$76644441 then
    Exception.Create('Not a mozilla registry file.');
  F.Read(MajorVersion, SizeOf(MajorVersion));
  F.Read(MinorVersion, SizeOf(MinorVersion));
  F.Read(NextAvail, SizeOf(NextAvail));
  F.Read(RootNode, SizeOf(RootNode));
end;

procedure TMozillaRegistry.Save(const FileName: string);
begin
  F:=TFileStream.Create(FileName, fmCreate);
  try
    SaveHeader;
    Root.SaveNode(F);
    Root.SaveChilds(F);
    NextAvail:=F.Position;
    F.Seek(NextAvailPosition, soFromBeginning);
    F.Write(NextAvail, SizeOf(NextAvail));
    RootNode:=Root.Location;
    F.Write(RootNode, SizeOf(RootNode));
  finally
    F.Free;
  end;
end;

procedure TMozillaRegistry.SaveHeader;
var
  b: byte;
  i: integer;
begin
  F.Write(Header, SizeOf(Header));
  F.Write(MajorVersion, SizeOf(MajorVersion));
  F.Write(MinorVersion, SizeOf(MinorVersion));
  NextAvailPosition:=F.Position;
  F.Write(NextAvail, SizeOf(NextAvail)); //Prozatim nevim, az ho budu vedet tak prepisu
  F.Write(RootNode, SizeOf(RootNode)); //Prozatim nevim, az ho budu vedet tak prepisu
  b:=0;
  for i:=F.Position to 127 do  //Reserver space
    F.Write(b, 1);
end;

end.
